@@ -9,7 +9,14 @@
 /* die with no strerror */
 #define DIE_N(msg) fprintf(stderr, "%s\n", msg); catchsig(1);
 
+#if HAVE_ADNS_H
+/* resolver master state */
+adns_state adns;
+#endif
+
+/* global stats */
 struct gstat gstats;
+
 time_t now;
 
 struct itemlist *items = NULL;
@@ -20,7 +27,7 @@ Timed_Circle *tc;
 Hits_Circle *hc;
 
 map *um, /* urlmap */
-//unused *im, /* ipmap */
+    *im, /* ipmap */
     *hm, /* hostmap */
     *rm, /* referrermap */
     *fm; /* filemap */
@@ -29,12 +36,14 @@ struct config cf;
 
 /* inputs */
 struct input *in;
-struct input_fd_table *ift;
 
 WINDOW *win;
 
-#if HAVE_KQUEUE
+#if (POLLING_METHOD == USING_KQUEUE)
 int kq;
+#elif (POLLING_METHOD == USING_FAM)
+/* master fd for talking to FAM */
+FAMConnection fc;
 #endif
 
 int main(int argc, char *argv[])
@@ -44,21 +53,24 @@ int main(int argc, char *argv[])
 	time_t last_display = 0;
 	struct logbits lb;
 	struct input *curfile;
-	struct input_fd_table *iftp;
 	bool seen_h = false, seen_t = false;
 	extern char *optarg;
-#if HAVE_KQUEUE
+#if (POLLING_METHOD == USING_KQUEUE)
+	/* we have and are using kqueue */
 	struct timespec stv;
-	stv.tv_sec = 1; stv.tv_nsec = 0;
 	struct kevent *ev, *evptr;
-#else
+#elif (POLLING_METHOD == USING_FAM)
+	/* we have and are using FAM */
+	int fam_fd;
+	FAMEvent fe;
+	struct timeval tv;
+	tv.tv_sec = 1; tv.tv_usec = 0;
+	fd_set readfds;
+
+#else /* stat() then */
 	struct stat sb;
-#endif
-
-	int udelay = 500000; /* half a second */
-
-#if HAVE_PTHREAD
-	pthread_t key_thread; /* monitors keyboard */
+	struct timeval tv;
+	tv.tv_sec = 1; tv.tv_usec = 0;
 #endif
 
 	LogParser *p;
@@ -76,8 +88,10 @@ int main(int argc, char *argv[])
 	cf.circle_size = DEFAULT_CIRCLE_SIZE;
 	cf.circle_mode = DEFAULT_CIRCLE_MODE;
 	cf.sort = DEFAULT_SORT;
+	cf.retcodes_sort = DEFAULT_RETCODES_SORT;
 	cf.refresh_delay = DEFAULT_REFRESH_DELAY;
 	cf.display_mode = DEFAULT_DISPLAY_MODE;
+	cf.numbers_mode = DEFAULT_NUMBERS_MODE;
 	cf.detail_display_urls = true;
 	cf.detail_display_hosts = true;
 	cf.detail_display_refs = true;
@@ -87,6 +101,7 @@ int main(int argc, char *argv[])
 	cf.keep_segments = 0;
 	cf.preserve_ref_protocol = 0;
 	cf.do_immed_display = false;
+	cf.do_resolving = false;
 	cf.urlfilter = new Filter();
 	cf.reffilter = new Filter();
 	cf.hostfilter = new Filter();
@@ -95,21 +110,33 @@ int main(int argc, char *argv[])
 	/* support MAX_INPUT_FILES input files */
 	in = (struct input *)calloc(MAX_INPUT_FILES, sizeof(struct input));
 
-	ift = (struct input_fd_table *)
-	    calloc(MAX_INPUT_FILES + 10, sizeof(struct input_fd_table));
-
-#if HAVE_KQUEUE /* then create kqueue {{{ */
+#if (POLLING_METHOD == USING_KQUEUE) /* then create kqueue {{{ */
 	if ((kq = kqueue()) < 0)
 	{
 		DIE("cannot create kqueue");
+		exit(1);
 	}
 
 	/* space in ev for all inputs */
 	ev = (struct kevent *)calloc(MAX_INPUT_FILES, sizeof(struct kevent));
-#endif /* }}} */
+/* }}} */
+#elif (POLLING_METHOD == USING_FAM) /* open FAM connection */
+	if (FAMOpen(&fc))
+	{
+		DIE("cannot connect to fam");
+		exit(1);
+	}
+
+	fam_fd = FAMCONNECTION_GETFD(&fc);
+#endif
+
+#if HAVE_ADNS_H
+	/* open adns connection */
+	adns_init(&adns, adns_if_noerrprint, 0);
+#endif
 
 	/* process commandline {{{ */
-	while ((ch = getopt(argc, argv, "f:H:T:hqls:pr:")) != -1)
+	while ((ch = getopt(argc, argv, "f:H:T:hqlrs:pd:")) != -1)
 	{
 		switch(ch)
 		{
@@ -150,6 +177,10 @@ int main(int argc, char *argv[])
 				cf.lowercase_urls = true;
 				break;
 
+			case 'r':
+				cf.do_resolving = true;
+				break;
+
 			case 's':
 				x = atoi(optarg);
 				if (x > 0)
@@ -160,7 +191,7 @@ int main(int argc, char *argv[])
 				cf.preserve_ref_protocol = 1;
 				break;
 
-			case 'r':
+			case 'd':
 				x = atoi(optarg);
 				if (x > 0)
 					cf.refresh_delay = x;
@@ -186,12 +217,14 @@ int main(int argc, char *argv[])
 			fprintf(stderr, "opening %s: %s\n",
 			        DEFAULT_LOGFILE, strerror(errno));
 			DIE_N("No input files could be opened");
+			exit(1);
 		}
 	}
 
 	if (seen_t && seen_h)
 	{
 		DIE_N("-T and -H are mutually exclusive. Specify only one.");
+		exit(1);
 	}
 
 
@@ -230,8 +263,8 @@ int main(int argc, char *argv[])
 	rm->create(cf.circle_size);
 
 	/* ip string -> ip hash map */
-//	im = new map; 
-//	im->create(cf.circle_size);
+	im = new map; 
+	im->create(cf.circle_size);
 
 	/* host string -> host hash map */
 	hm = new map; 
@@ -251,12 +284,7 @@ int main(int argc, char *argv[])
 	noecho();
 	keypad(win, true);
 
-#if HAVE_PTHREAD
-	nodelay(win, false);
-	pthread_create(&key_thread, NULL, &key_processor, NULL);
-#else
 	nodelay(win, true);
-#endif
 
 	for( ;; )
 	{
@@ -295,10 +323,10 @@ int main(int argc, char *argv[])
 			}
 		}
 
-#if !HAVE_PTHREAD
-		/* check for keypresses without threads */
-		if ((ch = wgetch(win)) != ERR)
-			read_key(ch);
+
+#if HAVE_ADNS_H
+		/* see if adns has got anything for us */
+		collect_dns_responses();
 #endif
 
 		/* see if it's time to update.. */
@@ -309,21 +337,20 @@ int main(int argc, char *argv[])
 		}
 
 		/* start of for() loop within these folds.. */
-#if HAVE_KQUEUE /* {{{ */
-		stv.tv_sec = 1; stv.tv_nsec = 0;
+#if (POLLING_METHOD == USING_KQUEUE) /* {{{ */
+		/* every 1/10th of a second */
+		stv.tv_sec = 0; stv.tv_nsec = 10000000;
 		x = kevent(kq, NULL, 0, ev, MAX_INPUT_FILES, &stv);
-		if (x == 0) continue; /* timeout, nothing happened */
+//		if (x == 0) continue; /* timeout, nothing happened */
 		if (x < 0) break; /* error */
 
 		for(evptr = ev ; evptr->filter ; evptr++)
 		{
-			/* set up pointer to input element for this fd */
-			iftp = &ift[evptr->ident];
-
-			/* now we can see which input element we're on */
-			curfile = &in[iftp->element];
-			
-			fd = evptr->ident;
+			/* udata contains a pointer to the struct input
+			** array element for this fd
+			*/
+			curfile = (struct input *)evptr->udata;
+			fd = curfile->fd; /* or evptr->ident will work */
 
 			/* see if we can deduce what happened */
 			if (((evptr->fflags & NOTE_DELETE) == NOTE_DELETE) ||
@@ -342,14 +369,37 @@ int main(int argc, char *argv[])
 				//read_amt = ev.data; /* we don't use this */
 			}
 /* }}} */
-#else /* fallback to stat() {{{*/
+#elif (POLLING_METHOD == USING_FAM) /* {{{ */
+		FD_ZERO(&readfds);
+		FD_SET(fam_fd, &readfds);
 
-		/* just sleep for a while */
-		usleep(udelay);
+		/* 1/10th of a second */
+		tv.tv_sec = 0;
+		tv.tv_usec = 100000;
 
-		/* when we have more than one file..
-		 * we'll have to check them all
-		*/
+		/* perform a select() on the fam filedescriptor */
+		select(fam_fd + 1, &readfds, NULL, NULL, &tv);
+
+		/* if FAM has nothing to report, loop around */
+		if (!(FD_ISSET(fam_fd, &readfds)))
+			continue;
+
+		while (FAMPending(&fc) == 1 && FAMNextEvent(&fc, &fe))
+		{
+			/* fe.userdata is struct input * element for this
+			** file; see new_file() for how it's set
+			*/
+			curfile = (struct input *)fe.userdata;
+			fd = curfile->fd;
+			
+/* }}} */
+#else /* fallback to stat() {{{ */
+
+		/* 1/10th of a second */
+		tv.tv_sec = 0; tv.tv_usec = 100000;
+		select(NULL, NULL, NULL, NULL, &tv);
+
+		/* check every file */
 		for (x = 0 ; x < MAX_INPUT_FILES ; ++x)
 		{
 			curfile = &in[x];
@@ -446,7 +496,13 @@ int main(int argc, char *argv[])
 
 			} /* while ((nl - buf) < buflen && nl) */
 		} /* for(evptr = ev ; evptr->filter ; evptr++) */
+		  /* while (FAMPending(&fc) == 1 && FAMNextEvent(&fc, &fe)) */
 		  /* for (i = 0 ; i < MAX_INPUT_FILES ; ++i) */
+
+		/* check for keypresses */
+		if ((ch = wgetch(win)) != ERR)
+			read_key(ch);
+
 	} /* for( ;; ) */
 
 	endwin();
@@ -472,25 +528,12 @@ int recordstats(struct logbits l) /* {{{ */
 	return 0;
 } /* }}} */
 
-#if HAVE_PTHREAD
-/* key_processor, monitors keyboard if we have threads {{{ */
-void *key_processor(void *)
-{
-	/* sit in getch() and pass to read_key when we get something */
-	int ch;
-
-	for( ;; )
-	{
-		ch = getch();
-		read_key(ch);
-	}
-} /* }}} */
-#endif
-
 int read_key(int ch) /* {{{ */
 {
-#define SUBMENU_SORT_TITLE "sort by.."
-#define SUBMENU_SORT_BANNER "r) REQUESTS  R) REQS/SEC  b) BYTES  B) BYTES/SEC"
+#define SUBMENU_SORT_HB_TITLE "sort by.."
+#define SUBMENU_SORT_HB_BANNER "r) REQUESTS  R) REQS/SEC  b) BYTES  B) BYTES/SEC"
+#define SUBMENU_SORT_RC_TITLE "sort by.."
+#define SUBMENU_SORT_RC_BANNER "2) 2xx   3) 3xx   4) 4xx   5) 5xx"
 
 #define SUBMENU_DISP_TITLE "toggle subdisplay.."
 #define SUBMENU_DISP_BANNER "u) URLS  r) REFERRERS  h) HOSTS"
@@ -507,14 +550,34 @@ int read_key(int ch) /* {{{ */
 	if (cf.in_submenu == SUBMENU_NONE) /* {{{ */
 	    switch(ch)
 	{
-		case 's': /* sort .. submenu */
-			cf.in_submenu = SUBMENU_SORT;
+		case 's': /* sort .. submenus */
 			cf.in_submenu_time = now;
 			cf.in_submenu_stay = false;
 
-			/* display banner to aid further presses */
-			display_submenu_banner(SUBMENU_SORT_TITLE,
-			    sizeof(SUBMENU_SORT_TITLE), SUBMENU_SORT_BANNER);
+			switch(cf.numbers_mode)
+			{
+				default:
+				case NUMBERS_HITS_BYTES:
+					cf.in_submenu = SUBMENU_SORT_HB;
+					/* display banner to aid further
+					 * presses */
+					display_submenu_banner(
+					    SUBMENU_SORT_HB_TITLE,
+					    sizeof(SUBMENU_SORT_HB_TITLE),
+					    SUBMENU_SORT_HB_BANNER);
+					break;
+
+				case NUMBERS_RETCODES:
+					cf.in_submenu = SUBMENU_SORT_RC;
+					/* display banner to aid further
+					 * presses */
+					display_submenu_banner(
+					    SUBMENU_SORT_RC_TITLE,
+					    sizeof(SUBMENU_SORT_RC_TITLE),
+					    SUBMENU_SORT_RC_BANNER);
+					break;
+			}
+
 			break;
 
 		case 't': /* toggle displays */
@@ -602,6 +665,19 @@ int read_key(int ch) /* {{{ */
 			cf.do_immed_display = true;
 			break;
 
+		case 'n': /* numbers mode; hits/bytes or returncodes */
+			switch(cf.numbers_mode)
+			{
+				case NUMBERS_HITS_BYTES:
+					cf.numbers_mode = NUMBERS_RETCODES;
+					break;
+				case NUMBERS_RETCODES:
+					cf.numbers_mode = NUMBERS_HITS_BYTES;
+					break;
+			}
+			cf.do_immed_display = true;
+			break;
+
 		case KEY_UP:
 			if (cf.display_mode != DISPLAY_DETAIL)
 			{
@@ -662,7 +738,8 @@ int read_key(int ch) /* {{{ */
 			cf.exit = true;
 			break;
 	} /* }}} */
-	else if (cf.in_submenu == SUBMENU_SORT) /* {{{ */
+
+	else if (cf.in_submenu == SUBMENU_SORT_HB) /* {{{ */
 	{
 		switch(ch)
 		{
@@ -687,6 +764,28 @@ int read_key(int ch) /* {{{ */
 		clear_submenu_banner();
 		cf.do_immed_display = true;
 	} /* }}} */
+	else if (cf.in_submenu == SUBMENU_SORT_RC) /* {{{ */
+	{
+		switch(ch)
+		{
+			case '2':
+				cf.retcodes_sort = SORT_2XX;
+				break;
+			case '3':
+				cf.retcodes_sort = SORT_3XX;
+				break;
+			case '4':
+				cf.retcodes_sort = SORT_4XX;
+				break;
+			case '5':
+				cf.retcodes_sort = SORT_5XX;
+				break;
+		}
+		cf.in_submenu = SUBMENU_NONE;
+		clear_submenu_banner();
+		cf.do_immed_display = true;
+	} /* }}} */
+
 	else if (cf.in_submenu == SUBMENU_DISP) /* {{{ */
 	{
 		switch(ch)
@@ -859,11 +958,24 @@ int new_file(char *filename, bool do_seek_to_end) /* {{{ */
 	int i, fd, input_element = -1;
 	struct stat sb;
 	struct input *this_file;
+	char realfile[MAXPATHLEN];
 
-#if HAVE_KQUEUE
+	/* if realpath cannot resolve the file, give up */
+	if ((realpath(filename, realfile) == NULL))
+		return -1;
+
+#if (POLLING_METHOD == USING_KQUEUE)
 	struct kevent kev;
 #endif
 
+	/* stat the item; ensure it's a file and get an inode */
+	if (stat(realfile, &sb) == -1)
+		return -1;
+
+	/* we can't tail anything except a file */
+//	if (sb.st_mode != S_IFREG)
+//		return -1;
+
 	/* choose where to put this input in the struct */
 	/* if it has an existing slot, re-use */
 	for(i = 0 ; i < cf.input_count ; ++i)
@@ -887,40 +999,40 @@ int new_file(char *filename, bool do_seek_to_end) /* {{{ */
 		input_element = cf.input_count;
 	}
 
-	if ((fd = open(filename, O_RDONLY)) == -1)
+	if ((fd = open(realfile, O_RDONLY)) == -1)
 		return -1;
 
-	/* set up ift to point at that */
-	ift[fd].element = input_element;
-
 	this_file = &in[input_element];
 	this_file->fd = fd;
 
-	/* now fstat() should work and give us an inode */
-	if (fstat(fd, &sb) == -1)
-	{
-		return -1;
-	}
-
 	if (do_seek_to_end) lseek(fd, 0, SEEK_END);
 
 	if (this_file->filename) free(this_file->filename);
 
 	this_file->inode = sb.st_ino;
-	this_file->filename = strdup(filename);
+	this_file->filename = strdup(realfile);
 	this_file->lastreq = now;
 	this_file->type = LOG_COMMON; /* assumption */
 	this_file->open = true;
 
-#if HAVE_KQUEUE
+#if (POLLING_METHOD == USING_KQUEUE)
 	/* add into kqueue */
+#ifdef __NetBSD__
 	EV_SET(&kev, fd, EVFILT_VNODE,
 	    EV_ADD | EV_ENABLE | EV_CLEAR,
-	    NOTE_WRITE | NOTE_DELETE | NOTE_RENAME, 0, 0);
+	    NOTE_WRITE | NOTE_DELETE | NOTE_RENAME, 0, (intptr_t)this_file);
+#else
+	EV_SET(&kev, fd, EVFILT_VNODE,
+	    EV_ADD | EV_ENABLE | EV_CLEAR,
+	    NOTE_WRITE | NOTE_DELETE | NOTE_RENAME, 0, this_file);
+#endif
 	if (kevent(kq, &kev, 1, NULL, 0, NULL) < 0)
 	{
 		DIE("cannot create kevent");
 	}
+#elif (POLLING_METHOD == USING_FAM)
+	/* add into FAM */
+	FAMMonitorFile(&fc, realfile, &this_file->famreq, this_file);
 #endif
 
 	return 0;
@@ -934,23 +1046,24 @@ void usage(void) /* {{{ */
 	    "  -f logfile  open logfile (assumed common/combined) [%s]\n"
 	    "              (repeat option for more than one source)\n"
 	    "\n"
-	    "URL/referrer munging options:\n"
+	    "URL/host/referrer munging options:\n"
 	    "  -q          keep query strings [%s]\n"
 	    "  -l          lowercase all URLs [%s]\n"
 	    "  -s num      keep num path segments of URL [all]\n"
 	    "  -p          preserve protocol at front of referrers [%s]\n"
+	    "  -r          resolve hostnames/IPs into each other [%s]\n"
 	    "\n"
 	    "Stats options:\n"
 	    "  Supply up to one of the following two. default: [-%c %d]\n"
 	    "  -H hits     remember stats for this many hits\n"
 	    "  -T secs     remember stats for this many seconds\n"
 	    "\n"
-	    "  -r secs     refresh delay in seconds [%d]\n"
+	    "  -d secs     refresh delay in seconds [%d]\n"
 	    "\n"
 	    "  -h          this help\n"
 	    "\n"
-	    "Compile Options: %cHAVE_KQUEUE %cENABLE_PCRE\n"
-	    "\n"
+	    "Compile Options: %cHAVE_KQUEUE %cHAVE_FAM %cENABLE_PCRE\n"
+	    "Polling Method: %s\n"
 	    ,
 	    PACKAGE_VERSION,
 	    DEFAULT_LOGFILE,
@@ -958,6 +1071,7 @@ void usage(void) /* {{{ */
 	    cf.keep_querystring ? "yes" : "no",
 	    cf.lowercase_urls ? "yes" : "no",
 	    cf.preserve_ref_protocol ? "yes" : "no",
+	    cf.do_resolving ? "yes" : "no",
 	    cf.circle_mode, cf.circle_size,
 	    cf.refresh_delay,
 #if HAVE_KQUEUE /* {{{ */
@@ -965,11 +1079,22 @@ void usage(void) /* {{{ */
 #else
 	    '-',
 #endif /* }}} */
-#if HAVE_PCRE_H /* {{{ */
-	    '+'
+#if HAVE_FAM_H /* {{{ */
+	    '+',
 #else
-	    '-'
+	    '-',
 #endif /* }}} */
+#if HAVE_PCRE_H /* {{{ */
+	    '+',
+#else
+	    '-',
+#endif /* }}} */
+	    (POLLING_METHOD == USING_KQUEUE ? "kqueue" :
+	        (POLLING_METHOD == USING_FAM ? "fam" :
+	            "stat"
+		)
+	    )
+	    
 	    );
 
 	return;
